import numpy as np
from toolkit.het_block import HetBlock


def curly(V_V, D_V, O_V, V_i, D_i, O_i, Pi_full, ss, T):
    """
    V_V: Jacobian of V vs. V_(+1)
    D_V: Jacobian of D_(+1) vs. V_(+1)
    O_V: dict of Jacobians of all outputs (lowercase) vs. V_(+1)
    V_i: dict of Jacobians of V vs. all inputs i
    D_i: dict of Jacobians of D vs. all inputs i
    O_i: nested dict of Jacobians of all outputs vs. all inputs i
    Pi: full transition matrix
    ss: steady state (must have distribution D and outputs o)
    T: number of periods

    Note that all Jacobians must be 2-dimensional (flatten dimensions!)
    """

    inputs = list(V_i.keys())
    outputs = list(O_V.keys())

    N_D, N_V = D_V.shape

    curlyY = {o.upper(): {i: np.empty(T) for i in inputs} for o in outputs}
    curlyD = {i: np.empty((T, N_D)) for i in inputs}
    curlyP = {o.upper(): np.empty((T, N_D)) for o in outputs}
    curlyV = {i: np.empty((T, N_V)) for i in inputs}

    Dss = ss['D'].ravel()

    # first backward iteration to get curlyY, curlyD, curlyV
    for i in inputs:
        curlyD[i][0] = D_i[i]
        curlyV[i][0] = V_i[i]
        for o in outputs:
            curlyY[o.upper()][i][0] = np.vdot(O_i[o][i], Dss)

        for t in range(1, T):
            V_1 = curlyV[i][t - 1]
            curlyD[i][t] = D_V @ V_1
            curlyV[i][t] = V_V @ V_1
            for o in outputs:
                curlyY[o.upper()][i][t] = np.vdot(O_V[o] @ V_1, Dss)

    # now apply Markov matrix to get curlyPs
    for o in outputs:
        curlyP[o.upper()][0] = ss[o].ravel()
        for t in range(1, T):
            curlyP[o.upper()][t] = Pi_full @ curlyP[o.upper()][t - 1]

    return curlyY, curlyD, curlyP, curlyV


def jacobians(curlyY, curlyD, curlyP):
    J = {}
    for o in curlyP:
        o = o.upper()
        J[o] = {}
        for i in curlyD:
            T = curlyY[o][i].shape[0]
            F = HetBlock.build_F(curlyY[o][i], curlyD[i], curlyP[o][:T - 1])
            J[o][i] = HetBlock.J_from_F(F)
    return J
