import numpy as np
import jax.numpy as jnp
from jax import custom_jvp, jacfwd
from jax_utils import interpolate
from jax.ops import index_update


def cn(uc, w, eis, frisch, vphi):
    """Return optimal c, n as function of u'(c) given parameters"""
    return uc ** (-eis), (w * uc / vphi) ** frisch


def solve_cn(w, T, eis, frisch, vphi, uc_seed):
    uc = solve_uc(w, T, eis, frisch, vphi, uc_seed)
    return cn(uc, w, eis, frisch, vphi)


@jnp.vectorize
def netexp(log_uc, w, T, eis, frisch, vphi):
    """Return net expenditure as a function of log uc and its derivative."""
    c, n = cn(jnp.exp(log_uc), w, eis, frisch, vphi)
    ne = c - w * n - T

    # c and n have elasticities of -eis and frisch wrt log u'(c)
    c_loguc = -eis * c
    n_loguc = frisch * n
    netexp_loguc = c_loguc - w * n_loguc

    return ne, netexp_loguc


@custom_jvp
@jnp.vectorize
def solve_uc(w, T, eis, frisch, vphi, uc_seed):
    """Solve for optimal uc given in log uc space.

    max_{c, n} c**(1-1/eis) + vphi*n**(1+1/frisch) s.t. c = w*n + T
    """
    log_uc = jnp.log(uc_seed)
    for i in range(10):
        ne, ne_p = netexp(log_uc, w, T, eis, frisch, vphi)
        log_uc -= ne / ne_p

    return jnp.exp(log_uc)


@jnp.vectorize
def get_derivative(uc, w, T, eis, frisch, vphi):
    return jacfwd(netexp, argnums=(0, 1, 2, 3, 4, 5))(jnp.log(uc), w, T, eis, frisch, vphi)[0]


@solve_uc.defjvp
def f_jvp(primals, tangents):
    w, T, eis, frisch, vphi, uc_seed = primals
    w_dot, T_dot, eis_dot, frisch_dot, vphi_dot, _ = tangents
    uc = solve_uc(w, T, eis, frisch, vphi, uc_seed)
    netexp_log_uc, netexp_w, netexp_T, netexp_eis, netexp_frisch, netexp_vphi = get_derivative(
        uc, w, T, eis, frisch, vphi)

    tangent = -(netexp_w * w_dot + netexp_T * T_dot + netexp_eis * eis_dot
                + netexp_frisch * frisch_dot + netexp_vphi * vphi_dot) / netexp_log_uc * uc

    return uc, tangent


def backward_iteration(Va_p, Pi_p, a_grid, e_grid, T, w, r, beta, eis, frisch, vphi, c_seed):
    uc_nextgrid = (beta * Pi_p) @ Va_p
    return backward_iteration_endog(uc_nextgrid, a_grid, e_grid, T, w, r, eis, frisch, vphi, c_seed)


def backward_iteration_endog(uc_nextgrid, a_grid, e_grid, T, w, r,
                             eis, frisch, vphi, c_seed):
    ws = w * e_grid
    c_nextgrid, n_nextgrid = cn(uc_nextgrid, ws[:, jnp.newaxis], eis, frisch, vphi)

    lhs = c_nextgrid - ws[:, jnp.newaxis] * n_nextgrid + a_grid[jnp.newaxis, :] - T[:, jnp.newaxis]
    rhs = (1 + r) * a_grid
    c = interpolate(rhs, lhs, c_nextgrid)
    n = interpolate(rhs, lhs, n_nextgrid)

    a = rhs + ws[:, jnp.newaxis] * n + T[:, jnp.newaxis] - c
    iconst = np.nonzero(a < a_grid[0])
    a = jnp.maximum(a, a_grid[0])

    uc_seed = c_seed[iconst] ** (-1 / eis)
    c_const, n_const = solve_cn(jnp.take(ws, iconst[0]),
                                jnp.take(rhs, iconst[1]) + jnp.take(T, iconst[0]) - a_grid[0], eis, frisch, vphi,
                                uc_seed)

    c = index_update(c, iconst, c_const)
    n = index_update(n, iconst, n_const)

    Va = (1 + r) * c ** (-1 / eis)

    ns = e_grid[:, np.newaxis] * n

    return Va, a, c, n, ns


"""jax-compatible other blocks"""


def firm(Y, w, Z, pi, mu, kappa):
    L = Y / Z
    Div = Y - w * L - mu / (mu - 1) / (2 * kappa) * jnp.log(1 + pi) ** 2 * Y
    return L, Div


def monetary(pi_L, pi, rstar_L, phi):
    r = (1 + rstar_L + phi * pi_L) / (1 + pi) - 1
    return r


def fiscal(r, B):
    Tax = r * B
    return Tax


def nkpc(pi, pi_P, w, Z, Y, Y_P, r_P, mu, kappa):
    nkpc_res = kappa * (w / Z - 1 / mu) + Y_P / Y * jnp.log(1 + pi_P) / (1 + r_P) - jnp.log(1 + pi)
    return nkpc_res


def transfers(pi_e, Div, Tax, div_rule, tax_rule):
    div = Div / jnp.sum(pi_e * div_rule) * div_rule
    tax = Tax / jnp.sum(pi_e * tax_rule) * tax_rule
    T = div - tax
    return T
