import numpy as np
import jax.numpy as jnp
from functools import partial

"""Part 1: Functions that are directly compatible with jax"""


@partial(jnp.vectorize, signature='(k),(l),(l)->(k)')
def interpolate(x, xp, fp):
    i = jnp.clip(jnp.searchsorted(xp, x, side='right'), 1, len(xp) - 1)
    df = jnp.take(fp, i) - jnp.take(fp, i - 1)
    dx = jnp.take(xp, i) - jnp.take(xp, i - 1)
    delta = x - jnp.take(xp, i - 1)
    f = jnp.take(fp, i - 1) + delta * (df / dx)
    return f


@partial(jnp.vectorize, signature='(k),(k)->(k),(k)')
def interpolate_rule(x, xp):
    i = jnp.clip(jnp.searchsorted(xp, x, side='right'), 1, len(xp) - 1)
    dx = jnp.take(xp, i) - jnp.take(xp, i - 1)
    delta = x - jnp.take(xp, i - 1)
    return i - 1, 1 - (delta / dx)


@partial(jnp.vectorize, signature='(k),(k)->(k,k)')
def to_Pi(x_i, x_pi):
    n = len(x_i)
    Ns = jnp.tile(jnp.arange(n), (n, 1))
    return (x_pi[:, jnp.newaxis] * (Ns == x_i[:, jnp.newaxis])
            + (1 - x_pi[:, jnp.newaxis]) * (Ns == (x_i[:, jnp.newaxis] + 1)))


@partial(jnp.vectorize, signature='(k),(k),(k)->(k)')
def update_policy(D, a, a_grid):
    a_i, a_pi = interpolate_rule(a, a_grid)
    return to_Pi(a_i, a_pi).T @ D


def update_dist(D, Pi, a, a_grid):
    return Pi.T @ update_policy(D, a, a_grid)


"""Part 2: Functions that replace some jax functionality for large grids when workarounds above start to crash jax"""


def to_Pi_full(Pi, a, a_grid):
    N = a.size
    a_i, a_pi = interpolate_rule(a, a_grid)
    Pi_endog = to_Pi(a_i, a_pi)
    return ((Pi.T[:, :, jnp.newaxis, jnp.newaxis] * Pi_endog)
            .transpose([1, 2, 0, 3]).reshape(N, N))


@partial(jnp.vectorize, signature='(k),(k)->(k,k)')
def to_dPi(x_i, x_grid):
    """Derivative of Pi matrix with respect to policy function at initial point"""
    inv_grid = np.concatenate((1 / (x_grid[1:] - x_grid[:-1]), [0]))
    inv_grid_1 = np.concatenate(([0], 1 / (x_grid[1:] - x_grid[:-1])))
    n = len(x_i)
    Ns = jnp.tile(jnp.arange(n), (n, 1))
    return (-inv_grid * (Ns == x_i[:, jnp.newaxis])
            + inv_grid_1 * (Ns == (x_i[:, jnp.newaxis] + 1)))


def get_Dmid_a(D, a, a_grid):
    a_i, _ = interpolate_rule(a, a_grid)
    dPi = to_dPi(a_i, a_grid)
    return D[:, :, jnp.newaxis] * dPi
