from toolkit import utils
import copy


class Composite:
    """Composition of multiple operators, with last applied first."""

    def __init__(self, *args):
        self.ops = args

    def __matmul__(self, x):
        for op in reversed(self.ops):
            x = op @ x
        return x

    @property
    def T(self):
        return Composite(*[op.T for op in reversed(self.ops)])

    @property
    def shape(self):
        return self.ops[0].shape[0], self.ops[-1].shape[-1]


class Motion:
    """Represents steady state full transition matrix (including exog and endog)"""

    def __init__(self, Pi, a, a_grid):
        self.x_i, self.x_pi = utils.interpolate_coord_robust(a_grid, a)

        self.Pi = Pi
        self.Pi_T = Pi.T.copy()
        self.transpose = False

    def __matmul__(self, D):
        if self.transpose:
            return utils.forward_step_transpose_1d(D, self.Pi, self.x_i, self.x_pi)
        else:
            return utils.forward_step_1d(D, self.Pi_T, self.x_i, self.x_pi)

    @property
    def T(self):
        M = copy.copy(self)
        M.transpose = not self.transpose
        return M


class PolicyToDist:
    """Represents mapping from perturbations to policy to perturbations to next-period distribution"""

    def __init__(self, D, Pi, x, x_grid):
        x_i = utils.interpolate_coord_robust(x_grid, x)[0]
        self.space = x_grid[x_i + 1] - x_grid[x_i]
        self.Pi_T = Pi.T
        self.D = D
        self.x_i = x_i

    def __matmul__(self, dx):
        x_pi_shock = -dx / self.space
        return utils.forward_step_shock_1d(self.D, self.Pi_T, self.x_i, x_pi_shock)


class Reshape:
    """Reshape from flat vector to these coordinates"""

    def __init__(self, n1, n2):
        self.n1 = n1
        self.n2 = n2
        self.transpose = False

    def __matmul__(self, A):
        if self.transpose:
            return A.ravel()
        else:
            return A.reshape(self.n1, self.n2)

    @property
    def T(self):
        R = copy.copy(self)
        R.transpose = not self.transpose
        return R

    @property
    def shape(self):
        if self.transpose:
            return self.n1 * self.n2, (self.n1, self.n2)
        else:
            return (self.n1, self.n2), self.n1 * self.n2
