"""
One-asset HANK model

Model DAG:
blocks     = [household, firm, monetary, fiscal, nkpc, mkt_clearing]
unknowns   = [K]
targets    = [asset_mkt]
exogenous  = [Z]
"""

import numpy as np
from numba import guvectorize
from toolkit import utils
from toolkit import het_block as het
from toolkit.simple_block import simple

'''Part 1: HA block'''


def backward_iterate(Va_p, Vb_p, Pi_p, a_grid, b_grid, z_grid, e_grid, k_grid, beta, eis, rb, ra, chi0, chi1, chi2):
    # require that k is decreasing
    assert k_grid[1] < k_grid[0], 'kappas in k_grid must be decreasing!'

    # precompute Psi1(a', a) on grid of (a', a) for steps 3 and 5
    Psi1 = get_Psi_and_deriv(a_grid[:, np.newaxis],
                             a_grid[np.newaxis, :], ra, chi0, chi1, chi2)[1]

    # === STEP 2: Wb(z, b', a') and Wa(z, b', a') ===
    # (take discounted expectation of tomorrow's value function)
    Wb = matrix_times_first_dim(beta * Pi_p, Vb_p)
    Wa = matrix_times_first_dim(beta * Pi_p, Va_p)
    W_ratio = Wa / Wb

    # === STEP 3: a'(z, b', a) for UNCONSTRAINED ===

    # for each (z, b', a), linearly interpolate to find a' between gridpoints
    # satisfying optimality condition W_ratio == 1+Psi1
    i, pi = lhs_equals_rhs_interpolate(W_ratio, 1 + Psi1)

    # use same interpolation to get Wb and then c
    a_endo_unc = utils.apply_coord(i, pi, a_grid)
    c_endo_unc = utils.apply_coord(i, pi, Wb) ** (-eis)

    # === STEP 4: b'(z, b, a), a'(z, b, a) for UNCONSTRAINED ===

    # solve out budget constraint to get b(z, b', a)
    b_endo = (c_endo_unc + a_endo_unc + addouter(-z_grid, b_grid, -(1 + ra) * a_grid)
              + get_Psi_and_deriv(a_endo_unc, a_grid, ra, chi0, chi1, chi2)[0]) / (1 + rb)

    # interpolate this b' -> b mapping to get b -> b', so we have b'(z, b, a)
    # and also use interpolation to get a'(z, b, a)
    # (note utils.interpolate_coord and utils.apply_coord work on last axis,
    #  so we need to swap 'b' to the last axis, then back when done)
    i, pi = utils.interpolate_coord(b_endo.swapaxes(1, 2), b_grid)
    a_unc = utils.apply_coord(i, pi, a_endo_unc.swapaxes(1, 2)).swapaxes(1, 2)
    b_unc = utils.apply_coord(i, pi, b_grid).swapaxes(1, 2)

    # === STEP 5: a'(z, kappa, a) for CONSTRAINED ===

    # for each (z, kappa, a), linearly interpolate to find a' between gridpoints
    # satisfying optimality condition W_ratio/(1+kappa) == 1+Psi1, assuming b'=0
    lhs_con = W_ratio[:, 0:1, :] / (1 + k_grid[np.newaxis, :, np.newaxis])
    i, pi = lhs_equals_rhs_interpolate(lhs_con, 1 + Psi1)

    # use same interpolation to get Wb and then c
    a_endo_con = utils.apply_coord(i, pi, a_grid)
    c_endo_con = ((1 + k_grid[np.newaxis, :, np.newaxis]) ** (-eis)
                  * utils.apply_coord(i, pi, Wb[:, 0:1, :]) ** (-eis))

    # === STEP 6: a'(z, b, a) for CONSTRAINED ===

    # solve out budget constraint to get b(z, kappa, a), enforcing b'=0
    b_endo = (c_endo_con + a_endo_con
              + addouter(-z_grid, np.full(len(k_grid), b_grid[0]), -(1 + ra) * a_grid)
              + get_Psi_and_deriv(a_endo_con, a_grid, ra, chi0, chi1, chi2)[0]) / (1 + rb)

    # interpolate this kappa -> b mapping to get b -> kappa
    # then use the interpolated kappa to get a', so we have a'(z, b, a)
    # (utils.interpolate_y does this in one swoop, but since it works on last
    #  axis, we need to swap kappa to last axis, and then b back to middle when done)
    a_con = utils.interpolate_y(b_endo.swapaxes(1, 2), b_grid,
                                a_endo_con.swapaxes(1, 2)).swapaxes(1, 2)

    # === STEP 7: obtain policy functions and update derivatives of value function ===

    # combine unconstrained solution and constrained solution, choosing latter
    # when unconstrained goes below minimum b
    a, b = a_unc.copy(), b_unc.copy()
    b[b <= b_grid[0]] = b_grid[0]
    a[b <= b_grid[0]] = a_con[b <= b_grid[0]]

    # calculate adjustment cost and its derivative
    Psi, _, Psi2 = get_Psi_and_deriv(a, a_grid, ra, chi0, chi1, chi2)

    # solve out budget constraint to get consumption and marginal utility
    c = addouter(z_grid, (1 + rb) * b_grid, (1 + ra) * a_grid) - Psi - a - b
    uc = c ** (-1 / eis)

    # for GE wage Phillips curve we'll need endowment-weighted utility too
    u = e_grid[:, np.newaxis, np.newaxis] * uc

    # update derivatives of value function using envelope conditions
    Va = (1 + ra - Psi2) * uc
    Vb = (1 + rb) * uc

    return Va, Vb, a, b, c, u


household = het.HetBlock(backward_iterate, exogenous='Pi', policy=['b', 'a'], backward=['Vb', 'Va'])

"""Supporting functions for HA block"""


def get_Psi_and_deriv(ap, a, ra, chi0, chi1, chi2):
    """Adjustment cost Psi(ap, a) and its derivatives with respect to first argument (ap) and second argument (a)."""
    a_with_return = (1 + ra) * a
    a_change = ap - a_with_return
    abs_a_change = np.abs(a_change)
    sign_change = np.sign(a_change)

    adj_denominator = a_with_return + chi0
    core_factor = (abs_a_change / adj_denominator) ** (chi2 - 1)

    Psi = chi1 / chi2 * abs_a_change * core_factor
    Psi1 = chi1 * sign_change * core_factor
    Psi2 = -(1 + ra) * (Psi1 + (chi2 - 1) * Psi / adj_denominator)
    return Psi, Psi1, Psi2


def matrix_times_first_dim(A, X):
    """Take matrix A times vector X[:, i1, i2, i3, ... , in] separately
    for each i1, i2, i3, ..., in. Same output as A @ X if X is 1D or 2D"""
    # flatten all dimensions of X except first, then multiply, then restore shape
    return (A @ X.reshape(X.shape[0], -1)).reshape(X.shape)


def addouter(z, b, a):
    """Take outer sum of three arguments: result[i, j, k] = z[i] + b[j] + a[k]"""
    return z[:, np.newaxis, np.newaxis] + b[:, np.newaxis] + a


@guvectorize(['void(float64[:], float64[:,:], uint32[:], float64[:])'], '(ni),(ni,nj)->(nj),(nj)')
def lhs_equals_rhs_interpolate(lhs, rhs, iout, piout):
    """
    Given lhs (i) and rhs (i,j), for each j, find the i such that

    lhs[i] > rhs[i,j] and lhs[i+1] < rhs[i+1,j]
    
    i.e. where given j, lhs == rhs in between i and i+1.
    
    Also return the pi such that 

    pi*(lhs[i] - rhs[i,j]) + (1-pi)*(lhs[i+1] - rhs[i+1,j]) == 0

    i.e. such that the point at pi*i + (1-pi)*(i+1) satisfies lhs == rhs by linear interpolation.

    If lhs[0] < rhs[0,j] already, just return u=0 and pi=1.

    ***IMPORTANT: Assumes that solution i is monotonically increasing in j
    and that lhs - rhs is monotonically decreasing in i.***
    """
    ni, nj = rhs.shape
    assert len(lhs) == ni

    i = 0
    for j in range(nj):
        while True:
            if lhs[i] < rhs[i, j]:
                break
            elif i < nj - 1:
                i += 1
            else:
                break

        if i == 0:
            iout[j] = 0
            piout[j] = 1
        else:
            iout[j] = i - 1
            err_upper = rhs[i, j] - lhs[i]
            err_lower = rhs[i - 1, j] - lhs[i - 1]
            piout[j] = err_upper / (err_upper - err_lower)


'''Part 2: Simple blocks and hetinput'''


@simple
def pricing(pi, mc, r, Y, kappap, mup, markup):
    nkpc = kappap * (mc - 1 / mup) + Y(+1) / Y * np.log(1 + pi(+1)) / (1 + r(+1)) + markup - np.log(1 + pi)
    return nkpc


@simple
def arbitrage(div, p, r):
    equity = div(+1) + p(+1) - p * (1 + r(+1))
    return equity


@simple
def labor(Y, w, K, Z, alpha):
    N = (Y / Z / K(-1) ** alpha) ** (1 / (1 - alpha))
    mc = w * N / (1 - alpha) / Y
    return N, mc


@simple
def investment(Q, K, r, N, mc, Z, delta, epsI, alpha, rinv_shock):
    inv = (K / K(-1) - 1) / (delta * epsI) + 1 - Q
    val = alpha * Z(+1) * (N(+1) / K) ** (1 - alpha) * mc(+1) - (K(+1) / K -
                                                                 (1 - delta) + (K(+1) / K - 1) ** 2 / (
                                                                         2 * delta * epsI)) + K(+1) / K * Q(+1) - (
                  1 + r(+1) + rinv_shock) * Q
    return inv, val


@simple
def dividend(Y, w, N, K, pi, mup, kappap, delta, epsI):
    psip = mup / (mup - 1) / 2 / kappap * np.log(1 + pi) ** 2 * Y
    k_adjust = K(-1) * (K / K(-1) - 1) ** 2 / (2 * delta * epsI)
    I = K - (1 - delta) * K(-1) + k_adjust
    div = Y - w * N - I - psip
    return I, div


@simple
def taylor(rstar, pi, Y, phi, phi_y):
    Y_target = 1
    i = rstar + phi * pi + phi_y * (Y - Y_target)
    # i = rstar + phi * pi
    return i


@simple
def fiscal(r, w, N, G, Bg):
    tax = (r * Bg + G) / w / N
    return tax


@simple
def finance(i, p, pi, r, div, omega, pshare):
    rb = r - omega
    ra = pshare * (div + p) / p(-1) + (1 - pshare) * (1 + r) - 1
    fisher = 1 + i(-1) - (1 + r) * (1 + pi)
    return rb, ra, fisher


@simple
def wage(pi, w):
    piw = (1 + pi) * w / w(-1) - 1
    return piw


@simple
def union(piw, N, tax, w, U, kappaw, muw, vphi, frisch, beta, markup_w):
    wnkpc = kappaw * (vphi * N ** (1 + 1 / frisch) - (1 - tax) * w * N * U / muw) + beta * np.log(
        1 + piw(+1)) + markup_w - np.log(1 + piw)
    return wnkpc


@simple
def mkt_clearing(p, A, B, Bg):
    asset_mkt = p + Bg - B - A
    return asset_mkt


def income(e_grid, tax, w, N):
    z_grid = (1 - tax) * w * N * e_grid
    return z_grid


household_inc = household.attach_hetinput(income)


@simple
def ha2_nodag(piw, psiw, rb, ra, tax, i, psip, I, Q, w, N, K, div, p, pi, mc, r, Y, Z, G, rstar, U, A, B, k_adjust,
              Bg, phi, kappap, mup, alpha, delta, epsI, omega, pshare, muw, kappaw, vphi, frisch, beta):
    nkpc = kappap * (mc - 1 / mup) + Y(+1) / Y * np.log(1 + pi(+1)) / (1 + r(+1)) - np.log(1 + pi)
    equity = div(+1) + p(+1) - p * (1 + r(+1))
    res_N = (Y / Z / K(-1) ** alpha) ** (1 / (1 - alpha))
    res_mc = w * N / (1 - alpha) / Y
    inv = (K / K(-1) - 1) / (delta * epsI) + 1 - Q
    val = alpha * Z(+1) * (N(+1) / K) ** (1 - alpha) * mc(+1) - (K(+1) / K -
           (1 - delta) + (K(+1) / K - 1) ** 2 / (2 * delta * epsI)) + K(+1) / K * Q(+1) - (1 + r(+1)) * Q
    res_psip = mup / (mup - 1) / 2 / kappap * np.log(1 + pi) ** 2 * Y - psip
    res_k_adjust = K(-1) * (K / K(-1) - 1) ** 2 / (2 * delta * epsI) - k_adjust
    res_I = K - (1 - delta) * K(-1) + k_adjust - I
    res_div = Y - w * N - I - psip - div
    res_i = rstar + phi * pi - i
    res_tax = (r * Bg + G) / w / N - tax
    res_rb = r - omega - rb
    res_ra = pshare * (div + p) / p(-1) + (1 - pshare) * (1 + r) - 1 - ra
    fisher = 1 + i(-1) - (1 + r) * (1 + pi)
    res_piw = (1 + pi) * w / w(-1) - 1 - piw
    res_psiw = muw / (1 - muw) / 2 / kappaw * np.log(1 + piw) ** 2 * N - psiw
    wnkpc = kappaw * (vphi * N ** (1 + 1 / frisch) - (1 - tax) * w * N * U / muw) + beta * np.log(1 + piw(+1)) - np.log(
        1 + piw)
    asset_mkt = p + Bg - B - A
    return nkpc, equity, res_N, res_mc, inv, val, res_psip, res_k_adjust, res_I, res_div, res_i, res_tax, res_rb, res_ra, fisher, res_piw, res_psiw, wnkpc, asset_mkt


'''Part 3: Steady state'''


def hank_ss(beta_guess=0.976, vphi_guess=2.07, chi1_guess=6.5, r=0.0125, tot_wealth=14, K=10, delta=0.02, kappap=0.1,
            muw=1.1, Bh=1.04, Bg=2.8, G=0.2, eis=0.5, frisch=1, chi0=0.25, chi2=2, epsI=4, omega=0.005, kappaw=0.1,
            phi=1.5, nZ=3, nB=50, nA=70, nK=50, bmax=50, amax=4000, kmax=1, rho_z=0.966, sigma_z=0.92, noisy=True):
    """Solve steady state of full GE model. Calibrate (beta, vphi, chi1, alpha, mup, Z) to hit targets for
       (r, tot_wealth, Bh, K, Y=N=1).
    """
    # set up grid
    b_grid = utils.agrid(amax=bmax, n=nB)
    a_grid = utils.agrid(amax=amax, n=nA)
    k_grid = utils.agrid(amax=kmax, n=nK)[::-1].copy()
    e_grid, _, Pi = utils.markov_rouwenhorst(rho=rho_z, sigma=sigma_z, N=nZ)

    # solve analytically what we can
    I = delta * K
    mc = 1 - r * (tot_wealth - Bg - K)
    alpha = (r + delta) * K / mc
    mup = 1 / mc
    Z = K ** (-alpha)
    w = (1 - alpha) * mc
    tax = (r * Bg + G) / w
    div = 1 - w - I
    p = div / r
    ra = r
    rb = r - omega

    # figure out initializer
    z_grid = income(e_grid, tax, w, 1)
    Va = (0.6 + 1.1 * b_grid[:, np.newaxis] + a_grid) ** (-1 / eis) * np.ones((z_grid.shape[0], 1, 1))
    Vb = (0.5 + b_grid[:, np.newaxis] + 1.2 * a_grid) ** (-1 / eis) * np.ones((z_grid.shape[0], 1, 1))

    # residual function
    def res(x):
        beta_loc, vphi_loc, chi1_loc = x
        if beta_loc > 0.999 / (1 + r) or vphi_loc < 0.001 or chi1_loc < 0.5:
            raise ValueError('Clearly invalid inputs')
        out = household_inc.ss(Va=Va, Vb=Vb, Pi=Pi, a_grid=a_grid, b_grid=b_grid, N=1, tax=tax, w=w, e_grid=e_grid,
                               k_grid=k_grid, beta=beta_loc, eis=eis, rb=rb, ra=ra, chi0=chi0, chi1=chi1_loc, chi2=chi2)
        asset_mkt = out['A'] + out['B'] - p - Bg
        labor_mkt = vphi_loc - (1 - tax) * w * out['U'] / muw
        return np.array([asset_mkt, labor_mkt, out['B'] - Bh])

    # solve for beta, vphi, omega
    (beta, vphi, chi1), _ = utils.broyden_solver(res, np.array([beta_guess, vphi_guess, chi1_guess]), noisy=noisy)

    # extra evaluation to report variables
    ss = household_inc.ss(Va=Va, Vb=Vb, Pi=Pi, a_grid=a_grid, b_grid=b_grid, N=1, tax=tax, w=w, e_grid=e_grid,
                          k_grid=k_grid, beta=beta, eis=eis, rb=rb, ra=ra, chi0=chi0, chi1=chi1, chi2=chi2)

    # other things of interest
    pshare = p / (tot_wealth - Bh)

    # calculate aggregate adjustment cost and check Walras's law
    chi = get_Psi_and_deriv(ss['a'], a_grid, r, chi0, chi1, chi2)[0]
    Chi = np.vdot(ss['D'], chi)
    goods_mkt = ss['C'] + I + G + Chi + omega * ss['B'] - 1
    assert np.abs(goods_mkt) < 1E-7

    ss.update({'pi': 0, 'piw': 0, 'Q': 1, 'Y': 1, 'N': 1, 'mc': mc, 'K': K, 'Z': Z, 'I': I, 'w': w, 'tax': tax,
               'div': div, 'p': p, 'r': r, 'Bg': Bg, 'G': G, 'Chi': Chi, 'goods_mkt': goods_mkt, 'chi': chi, 'phi': phi,
               'beta': beta, 'vphi': vphi, 'omega': omega, 'alpha': alpha, 'delta': delta, 'mup': mup, 'muw': muw,
               'frisch': frisch, 'epsI': epsI, 'a_grid': a_grid, 'b_grid': b_grid, 'z_grid': z_grid, 'e_grid': e_grid,
               'k_grid': k_grid, 'Pi': Pi, 'kappap': kappap, 'kappaw': kappaw, 'pshare': pshare, 'rstar': r, 'i': r,
               'markup': 0, 'markup_w': 0, 'rinv_shock': 0, 'phi_y': 0, 'psip': 0, 'psiw': 0, 'k_adjust': 0})

    return ss


'''Find new steady state given Z'''


def ssmap(ss, r, Y):
    # evaluate production side
    capital_share = ss['alpha'] / ss['mup']
    K = (capital_share / (r + ss['delta'])) * Y
    N = (Y / ss['Z'] / K ** ss['alpha']) ** (1 / (1 - ss['alpha']))
    labor_share = (1 - ss['alpha']) / ss['mup']
    w = labor_share * Y / N
    I = ss['delta'] * K
    div = Y - w * N - I
    p = div / r

    # taxes
    tax = fiscal.ss(r, w, N, ss['G'], ss['Bg'])

    # finance
    rb = r - ss['omega']
    ra = r

    # now evaluate household
    out = household_inc.ss(Va=ss['Va'], Vb=ss['Vb'], Pi=ss['Pi'], a_grid=ss['a_grid'], b_grid=ss['b_grid'],
                           N=N, tax=tax, w=w, e_grid=ss['e_grid'], k_grid=ss['k_grid'], beta=ss['beta'], eis=ss['eis'],
                           rb=rb, ra=ra, chi0=ss['chi0'], chi1=ss['chi1'], chi2=ss['chi2'])

    # create residuals
    asset_mkt = out['A'] + out['B'] - p - ss['Bg']
    labor_mkt = ss['vphi'] * N ** (1 / ss['frisch']) - (1 - tax) * w * out['U'] / ss['muw']

    # other things of interest
    pshare = p / (p + ss['Bg'])
    chi = get_Psi_and_deriv(out['a'], ss['a_grid'], r, ss['chi0'], ss['chi1'], ss['chi2'])[0]
    Chi = np.vdot(out['D'], chi)
    goods_mkt = out['C'] + I + ss['G'] + Chi + ss['omega'] * out['B'] - Y

    return asset_mkt, labor_mkt, {**ss, **out, **dict(K=K, N=N, w=w, I=I, div=div, p=p, tax=tax, Y=Y, r=r,
                                                      rb=rb, ra=ra, pshare=pshare, Chi=Chi, goods_mkt=goods_mkt, i=r,
                                                      rstar=r)}


def hank_ss_Z(dZ, ss):
    # Define new value of productivity
    Znew = dZ * ss['Z']

    # Compute new equilibrium level of output and interest rate    
    ssalt = {**ss, 'Z': Znew}
    (r, Y), _ = utils.broyden_solver(lambda x: np.array(ssmap(ssalt, *x)[:2]), np.array([ss['r'], ss['Y']]),
                                     noisy=False)
    ssnew = ssmap(ssalt, r, Y)[2]

    # Check Walras law
    assert ssnew['goods_mkt'] < 1E-7

    return ssnew
