"""
Baseline model from Herbst and Schorfheide (2015)

Notes:
Measurement block is here for estimation but is not necessary to solve for IRF

Model DAG:
blocks     = [euler, nkpc, taylor, measurement]
unknowns   = [y, pi, R]
targets    = [yres, Rres, pires]
exogenous  = [epsR, g, z]
"""

from toolkit.simple_block import simple

'''Part 1: Blocks'''


@simple
def euler(y, R, pi, z, g, tau):
    yres = y(1) - 1 / tau * (R - pi(1) - z(1)) + g - g(1) - y
    return yres


@simple
def nkpc(pi, y, g, beta, kappa):
    pires = beta * pi(1) + kappa * (y - g) - pi
    return pires


@simple
def taylor(R, pi, y, g, epsR, rho_r, psi1, psi2):
    Rres = rho_r * R(-1) + (1 - rho_r) * psi1 * pi + (1 - rho_r) * psi2 * (y - g) + epsR - R
    return Rres


@simple
def measurement(y, pi, R, z):
    gdp = (y - y(-1) + z)
    infl = 4 * pi
    interest = 4 * R
    return gdp, infl, interest


'''Part 2: Steady state'''


def calibrate_ss():
    ss = dict()

    # Parameters
    ss['tau'] = 2.83
    ss['kappa'] = 0.78
    ss['psi1'] = 1.8
    ss['psi2'] = 0.63
    ss['rho_r'] = 0.77
    ss['rho_g'] = 0.98
    ss['rho_z'] = 0.88
    ss['sigma_r'] = 0.22
    ss['sigma_g'] = 0.71
    ss['sigma_z'] = 0.31

    # Steady state values
    ss['pi'] = 0
    ss['y'] = 0
    ss['R'] = 0

    # Shock process
    ss['g'] = 0
    ss['epsR'] = 0
    ss['z'] = 0

    # Measurement parameters
    ss['data_r'] = 0.42
    ss['data_pi'] = 3.3
    ss['data_gamma'] = 0.52
    ss['beta'] = 1 / (1 + ss['data_r'] / 400)

    return ss
